/* hgen+ */
/*
 * $Id: kk.c,v 1.5 2006/07/25 07:43:42 az Exp $
 * 
 * File:		kk.c
 * Date:		Thu Nov  3 14:40:58 2005
 * Author:		Alexander Zangerl (az)
 * 
 * Abstract:
 * Modifications:
 *	$Log: kk.c,v $
 *	Revision 1.5  2006/07/25 07:43:42  az
 *	include file is keyutils.h, no longer keyutil.h
 *
 *	Revision 1.4  2006/01/17 03:08:55  az
 *	changed failed deletion to not return an error code
 *
 *	Revision 1.3  2005/11/04 04:40:37  az
 *	changed xgetpass to not display a prompt if fed from pipe
 *
 *	Revision 1.2  2005/11/04 02:44:48  az
 *	changed get to request_key (kernel instantiation)
 *
 *	Revision 1.1  2005/11/03 05:45:09  az
 *	Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <asm/unistd.h>
#include <errno.h>
#include <termios.h>

#include "keyutils.h"

#define DIE(...) { fprintf(stderr,__VA_ARGS__);exit(1);}
#define USAGE  DIE("usage: \t%s {put|get|delete} ID\n\t%s list\n",argv[0],argv[0])

/* snarfed from quintuple-agent, adjusted a bit */
/* xgetpass - get a passphrase 
   Output PROMPT and then input the password with echo turned off.
   The password is stored in a new buffer */
char *xgetpass(char *prompt)
{
  char *buf, *bp;
  size_t len, bufsize = 1000;
  struct termios old, new;

  if (isatty(STDIN_FILENO)) 
  {
     if (tcgetattr(STDIN_FILENO, &old) != 0) 
     {
	perror("could not get terminal characteristics");
	return NULL;
     }
    new = old;
    new.c_lflag &= ~ECHO;
    if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &new) != 0) 
    {
       perror("could not turn echoing off");
       return NULL;
    }
    printf("%s", prompt);
  }
  buf = (char *)malloc(bufsize);
  if (!buf) 
  {
    fprintf(stderr, "could not allocate space\n");
    return NULL;
  }
  len = 0;
  while (1) 
  {
     bp = buf + len;
     if (!fgets(bp, bufsize - len, stdin)) {
	perror("error while reading");
	free(buf);
	return NULL;
     }
    len += strlen(bp);
    if (buf[len-1] == '\n')
      break;			/* we've read the entire line */
    bufsize *= 2;
    buf = (char *)realloc(buf, bufsize);
    if (!buf) {
       fprintf(stderr, "could not allocate space\n");
       return NULL;
    }
  }
  bp[len-1] = 0;
  if (isatty(STDIN_FILENO) && tcsetattr(STDIN_FILENO, TCSAFLUSH, &old) != 0)
     perror("could not turn echoing back on");
  putchar('\n');
  return buf;
}  


int main(int argc, char *argv[])
{
   if (argc<=1 || argc>3)
   {
      USAGE;
   }

   if (!strncasecmp(argv[1],"list",4))
   {
      key_serial_t *keylist,*p,k;
      int r,c;
      char *buffer;
      int type,dpos,dlen;
      uid_t uid; 
      gid_t gid;
      key_perm_t mode;

      c=keyctl_read_alloc(KEY_SPEC_USER_KEYRING,(void *)&keylist);
      if (c<0)
	 DIE("listing keyring failed: %s\n",strerror(errno));
      c/=sizeof(key_serial_t);
      p=keylist;
      while(c)
      {
	 k=*p++;
	 r=keyctl_describe_alloc(k,&buffer);
	 if (r<0)
	    DIE("key %d inaccessible: %s\n",k,strerror(errno));
	 sscanf(buffer, "%*[^;]%n;%u;%u;%x;%n%*[^;]%n", &type, &uid, 
		&gid, &mode,&dpos,&dlen);
	 if (dlen == -1)
	    DIE("description unparseable: %s\n",strerror(errno));
	 printf("%d: %s\n",k,buffer+dpos);
	 c--;
      }
   }
   else if (!strncasecmp(argv[1],"put",3))
   {
      int r;
      char *keyid=argv[2],prompt[1000],*payload;

      if (argc!=3)
	 USAGE;

      sprintf(prompt,"Enter secret to store under \"%s\": ",
	       keyid);
      payload=xgetpass(prompt);
      if (!payload)
	 DIE("could not read secret\n");
      r=add_key("user",keyid,payload,strlen(payload),KEY_SPEC_USER_KEYRING);
      if (r<0)
	 DIE("add key failed: %s\n",strerror(errno));
   }
   else if (!strncasecmp(argv[1],"get",3))   
   {
      key_serial_t kid;
      char *keyid=argv[2],*payload;
      int r;

      if (argc!=3)
	 USAGE;

      /* request this key from kernel (which will try to instantiate it
	 if it's not there) 
	 attach newly created key to user keyring*/
      kid=request_key("user",keyid,keyid,KEY_SPEC_USER_KEYRING);
      if (kid<0)
	 DIE("no key \"%s\" found: %s\n",keyid,strerror(errno));
      
      /* retrieve key */
      r=keyctl_read_alloc(kid,(void *)&payload);
      if (r<0)
	 DIE("key retrieval failed: %s\n",strerror(errno));

    if (isatty(STDOUT_FILENO))
	printf("secret available, but I won't print it on a tty\n");
    else
       puts(payload);
   }
   else if (!strncasecmp(argv[1],"delete",6))   
   {
      key_serial_t kid;
      char *keyid=argv[2];
      int r;

      if (argc!=3)
	 USAGE;

      /* find the key */
      kid=keyctl_search(KEY_SPEC_USER_KEYRING,"user",keyid,0);
      if (kid<0)
	 fprintf(stderr,"no key \"%s\" found: %s\n",keyid,strerror(errno));
      else
      {
	 /* unlink it */
	 r=keyctl_unlink(kid,KEY_SPEC_USER_KEYRING);
	 if (r<0)
	    DIE("unlink failed: %s\n",strerror(errno));
      }
   }
   else 
   {
      USAGE;
   }
   return 0;
}


